'''

'''
import time
from .base_command import BaseRSMQCommand


class SetQueueAttributesCommand(BaseRSMQCommand):
    '''
    Get Queue Attributes if does not exist
    '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'vt': {'required': False,
                     'value': None},
              'delay': {'required': False,
                        'value': None},
              'maxsize': {'required': False,
                          'value': None},
              'quiet': {'required': False,
                        'value': False}
              }

    def exec_command(self):
        ''' Exec Command '''
        now = int(time.time())

        queue_key = self.queue_key

        tx = self.client.pipeline(transaction=True)
        for param in ['vt', 'delay', 'maxsize']:
            value = self.param_get(param, default_value=None)
            if value is not None:
                tx.hset(queue_key, param, value)
        if tx:
            self.log.debug("Applying queue attribute changes")
            tx.hset(queue_key, "modified", now)
        else:
            self.log.debug("No queue attribute changes")
        _result = tx.execute()

        return self.parent.getQueueAttributes().qname(self.get_qname).execute()



'''
Python Redis Simple Queue Manager Consumer
'''


import logging
from threading import Thread
import time

from rsmq.cmd.exceptions import NoMessageInQueue

from .cmd import utils
from .cmd.exceptions import RedisSMQException
from .rsmq import RedisSMQ
from .rsmq import const


LOG = logging.getLogger(__name__)


class RedisSMQConsumer():
    '''
    RSMQ Consumer Worker
    '''
    # local parameters and their value
    LOCAL_PARAMS = {'retry_delay': 0,
                    'empty_queue_delay': 2.0,
                    'decode': True}

    class VisibilityTimeoutExtender(Thread):
        ''' Thread that keeps the visibility '''

        def __init__(self, consumer, message_id):
            ''' Initialize '''
            self.consumer = consumer
            self.rsqm = consumer.rsqm

            self.vt = consumer.vt
            self.message_id = message_id
            thread_name = "VTExtender:%s" % message_id
            super(RedisSMQConsumer.VisibilityTimeoutExtender,
                  self).__init__(name=thread_name)

            self.daemon = True
            self._request_stop = False

            # How often to check time
            self.interval = self.vt / 4

            self.next_extension = time.time() + self.vt / 2
            self.set_next_extension()

        def set_next_extension(self):
            ''' set next extension '''
            self.next_extension = time.time() + self.vt / 2

        def trace(self, msg, *args):
            ''' Passthrough for trace messages '''
            self.consumer.trace("VisibilityTimeoutExtender: %s" % msg, args)

        def extend(self):
            ''' extend visibility timeout '''
            try:
                self.trace("Extending visibility by %s seconds...", self.vt)
                self.rsqm.changeMessageVisibility(
                    vt=self.vt, id=self.message_id).execute()
                self.set_next_extension()
            except RedisSMQException as ex:
                LOG.warning(
                    "Failed to extend message visibility for %s: %s", self.message_id, ex)

        def stop(self):
            ''' Stop this thread '''
            self._request_stop = True

        def run(self):
            ''' Run'''
            while not self._request_stop:
                time.sleep(self.interval)
                if not self._request_stop and time.time() > self.next_extension:
                    self.extend()
                    self.set_next_extension()

    def __init__(self, qname, processor, **rsmq_params):
        '''
        initialize 

        Required Parameters:

        @param qname: queue name
        @param processor: processor for each item

        Optional Parameters: 
        @param retry_delay: amount of time, in seconds, before failed item is retried
        @param empty_queue_delay: (float) Amount of time, in seconds, to wait if no items in queue  

        Remaining args are passed to RedisSMQ()

        '''
        self._request_stop = False
        self.rsqm = RedisSMQ(qname=qname, **rsmq_params)
        self.qname = qname
        self.processor = processor

        self._trace = rsmq_params.get('trace', False)

        # Separate local params from kwargs
        self.params = {}
        for param, value in self.LOCAL_PARAMS.items():
            self.params[param] = rsmq_params.get(param, value)
            if param in rsmq_params:
                del rsmq_params[param]

        # get vt from kwargs, if set
        self.vt = rsmq_params.get('vt', None)
        # get vt from queue if not set in kwargs and queue exists
        self._get_vt()

    def _param(self, param, default_value=None):
        ''' get local param '''
        return self.params.get(param, self.LOCAL_PARAMS.get(param, default_value))

    @property
    def retry_delay(self):
        ''' retry delay '''
        return self._param('retry_delay')

    @property
    def empty_queue_delay(self):
        ''' empty_queue_delay '''
        return self._param('empty_queue_delay')

    @property
    def decode(self):
        ''' decode, if true, attempt to decode the output from JSON '''
        return self._param('decode')

    def _get_vt(self):
        ''' Get VT from the queue info if not set '''
        if self.vt is None:
            queue_info = self.rsqm.getQueueAttributes().qname(
                self.qname).exceptions(False).execute()
            if queue_info and 'vt' in queue_info:
                self.vt = int(queue_info.get('vt', const.VT_DEFAULT))
            else:
                self.vt = const.VT_DEFAULT

    def stop(self, wait=None):
        ''' Stop '''
        self._request_stop = True
        if wait:
            wait_until = time.time() + wait
            while not time.time() > wait_until:
                time.sleep(0.5)
                if self._request_stop is None:
                    break
        return self._request_stop is None

    def on_success(self, msg):
        ''' Run on success '''
        # delete the item
        if msg and 'id' in msg:
            self.trace("Processed message %s", msg['id'])
            self.rsqm.deleteMessage(qname=self.qname, id=msg['id']).execute()

    def on_failure(self, msg):
        ''' Run on success '''

        if msg and 'id' in msg:
            LOG.warning("Failed to process message %s", msg['id'])
            self.rsqm.changeMessageVisibility(vt=self.retry_delay,
                                              qname=self.qname,
                                              id=msg['id']).execute()

    def create_queue(self):
        ''' Create queue if it does not exists '''
        self.rsqm.createQueue(
            qname=self.qname, quiet=True).exceptions(False).execute()

    def trace(self, fmt, *args):
        ''' Print trace log messages for debugging, if enabled '''
        if self._trace is not False:
            LOG.debug(fmt, *args)

    def run(self):
        ''' main loop of the thread '''
        self.trace("Starting Queue Consumer for %s", self.qname)
        self.create_queue()
        while not self._request_stop:
            try:
                msg = self.rsqm.receiveMessage().execute()
                if msg and 'id' in msg:
                    extender = RedisSMQConsumer.VisibilityTimeoutExtender(
                        self, msg['id'])
                    extender.start()
                    if self.decode and 'message' in msg:
                        msg['message'] = utils.decode_message(msg['message'])
                    if self.processor(**msg):
                        self.on_success(msg)
                    else:
                        self.on_failure(msg)
                    extender.stop()
                else:
                    raise RedisSMQException(
                        "Invalid message in queue '%s': %s" % (self.qname, msg))
            except NoMessageInQueue:
                delay = self.empty_queue_delay
                self.trace("No message in queue, waiting %s seconds", delay)
                if delay:
                    time.sleep(delay)
            except RedisSMQException as ex:
                LOG.warning("Exception while processing queue `%s`: %s",
                            self.qname, ex)
        self._request_stop = None
        self.trace("Ended Queue Consumer for %s", self.qname)


class RedisSMQConsumerThread(RedisSMQConsumer, Thread):
    ''' Version of a RedisSMQConsumer implemented as a self-contained thread '''

    def __init__(self, qname, processor, **rsmq_params):
        ''' Constructor '''
        RedisSMQConsumer.__init__(self, qname, processor, **rsmq_params)
        Thread.__init__(self, name="RedisSMQConsumer:%s" % qname, daemon=True)



'''
Constant values
'''

# max length of queue name
QNAME_MAX_LEN = 64

QNAME_INVALID_CHARS_RE = r'[^A-Za-z0-9._-]'

# Suffix to append to the queue
QUEUE_SUFFUX = ":Q"

# suffix to append to the queue set
QUEUES = "QUEUES"

# minimum VT - seconds
VT_MIN = 0

# maximum VT - seconds
VT_MAX = 9999999

VT_DEFAULT = 30


# minimum DELAY - seconds
DELAY_MIN = 0

# maximum DELAY - seconds
DELAY_MAX = 9999999

DELAY_DEFAULT = 0


MAXSIZE_MIN = 1024
MAXSIZE_MAX = 65565
MAXSIZE_DEFAULT = MAXSIZE_MAX


SCRIPT_POPMESSAGE = 'local msg = redis.call("ZRANGEBYSCORE", KEYS[1], "-inf", KEYS[2], "LIMIT", "0", "1") if #msg == 0 then return {} end redis.call("HINCRBY", KEYS[1] .. ":Q", "totalrecv", 1) local mbody = redis.call("HGET", KEYS[1] .. ":Q", msg[1]) local rc = redis.call("HINCRBY", KEYS[1] .. ":Q", msg[1] .. ":rc", 1) local o = {msg[1], mbody, rc} if rc==1 then table.insert(o, KEYS[2]) else local fr = redis.call("HGET", KEYS[1] .. ":Q", msg[1] .. ":fr") table.insert(o, fr) end redis.call("ZREM", KEYS[1], msg[1]) redis.call("HDEL", KEYS[1] .. ":Q", msg[1], msg[1] .. ":rc", msg[1] .. ":fr") return o'
SCRIPT_RECEIVEMESSAGE = 'local msg = redis.call("ZRANGEBYSCORE", KEYS[1], "-inf", KEYS[2], "LIMIT", "0", "1") if #msg == 0 then return {} end redis.call("ZADD", KEYS[1], KEYS[3], msg[1]) redis.call("HINCRBY", KEYS[1] .. ":Q", "totalrecv", 1) local mbody = redis.call("HGET", KEYS[1] .. ":Q", msg[1]) local rc = redis.call("HINCRBY", KEYS[1] .. ":Q", msg[1] .. ":rc", 1) local o = {msg[1], mbody, rc} if rc==1 then redis.call("HSET", KEYS[1] .. ":Q", msg[1] .. ":fr", KEYS[2]) table.insert(o, KEYS[2]) else local fr = redis.call("HGET", KEYS[1] .. ":Q", msg[1] .. ":fr") table.insert(o, fr) end return o'
SCRIPT_CHANGEMESSAGEVISIBILITY = 'local msg = redis.call("ZSCORE", KEYS[1], KEYS[2]) if not msg then return 0 end redis.call("ZADD", KEYS[1], KEYS[3], KEYS[2]) return 1'



'''

'''

from .base_command import BaseRSMQCommand


class ListQueuesCommand(BaseRSMQCommand):
    '''
    List Queues.

    On execution returns a set of queues already existing on the redis in this namespace
    '''

    PARAMS = {'quiet': {'required': False,
                        'value': False}}

    def exec_command(self):
        ''' Exec Command '''
        client = self.client
        ret = client.smembers(self.queue_set)
        return ret



'''
Base Command for all the commands
'''

import logging
import re

from .. import const
from .exceptions import CommandNotImplementedException
from .exceptions import InvalidParameterValue
from .exceptions import QueueDoesNotExist
from .exceptions import RedisSMQException
from .utils import validate_int


# REGEX matching invalid QNAME characters
QNAME_INVALID_RE = re.compile(const.QNAME_INVALID_CHARS_RE)


class BaseRSMQCommand():
    ''' Base for all RQMS commands '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'quiet': {'required': False,
                        'value': False}
              }

    def __init__(self, rsmq, **kwargs):
        self.log = logging.getLogger('rsmq.cmd.%s' % self.__class__.__name__)
        self.parent = rsmq
        self._params = {}
        # Load Defaults
        for name, value in self._param_defaults().items():
            self._params[name] = value
        # load args
        for name, value in kwargs.items():
            self._set_param(name, value)
        self.__exceptions = None

    @property
    def popMessageSha1(self):
        ''' popMessageSha1 '''
        return self.parent.popMessageSha1

    @property
    def receiveMessageSha1(self):
        ''' receiveMessageSha1 '''
        return self.parent.receiveMessageSha1

    @property
    def changeMessageVisibilitySha1(self):
        ''' changeMessageVisibilitySha1 '''
        return self.parent.changeMessageVisibilitySha1

    @property
    def _exceptions(self):
        ''' Returns true if exceptions are enabled'''
        if self.__exceptions is None:
            return self.parent.options.get('exceptions', True)
        return self.__exceptions

    def exceptions(self, enabled=True):
        ''' Control exceptions '''
        self.__exceptions = enabled != False
        return self

    @property
    def client(self):
        ''' Get Redis client '''
        return self.parent.client

    def _param_defaults(self):
        ''' Get dict of default parameters and their values '''
        defaults = {}
        for name, value in self.PARAMS.items():
            defaults[name] = value.get('value', None)
        return defaults

    def _required_params(self):
        ''' List of parameters that are required '''
        return [param
                for param, definition in self.PARAMS.items()
                if definition.get('required', False) == True]

    def config(self, key, default_value):
        ''' Get value from global config '''
        return self.parent.options.get(key, default_value)

    @property
    def namespace(self):
        ''' Get namespace '''
        namespace = self.config('ns', '')
        if namespace:
            return namespace + ':'
        return ''

    @property
    def queue_base(self):
        ''' Get base name of the queue '''
        return self.namespace + self.get_qname

    @property
    def queue_key(self):
        ''' Get Full queue name '''
        return self.queue_base + const.QUEUE_SUFFUX

    @property
    def queue_set(self):
        ''' Get Full Queue Set '''
        return self.namespace + const.QUEUES

    def param_get(self, param, default_value=None):
        ''' Get parameter by name '''
        return self._params.get(param, self.PARAMS[param].get('default', default_value))

    def __getattr__(self, name):
        ''' 
        Create setters for parameters on the fly 
        '''
        if name in self.PARAMS:
            def setter(value):
                if self._validate_param(name, value):
                    self._params[name] = value
                elif self._exceptions:
                    raise InvalidParameterValue(name, value)
                elif self.get_quiet is not True:
                    self.log.info("Invalid value for '%s': '%s'", name, value)
                return self
            return setter
        elif name.startswith('get_'):
            param = name[4:]
            if param in self.PARAMS:
                return self.param_get(param, default_value=None)

        raise AttributeError("'%s' object has no attribute '%s'" %
                             (self.__class__.__name__, name))

    def _validate_param(self, name, value):
        ''' Validate parameter value '''
        if hasattr(self, "_validate_%s" % name):
            validator = getattr(self, "_validate_%s" % name)
            return validator(value)
        return self._default_validator(name, value)

    def _default_validator(self, name, value):
        ''' Default Validator '''
        if not name:
            return False
        if not value:
            return False
        return True

    def _validate_qname(self, qname):
        ''' Validate Queue Name '''
        # Check we have a value at all
        if not qname:
            return False
        # check length
        if len(qname) > const.QNAME_MAX_LEN:
            return False
        # check if we have invalid characters
        if ':' in qname:
            print('Detected :')
            return False
        return True

    def _validate_vt(self, vt):
        ''' Validate visibility timeout parameter '''
        return validate_int(vt, const.VT_MIN, const.VT_MAX, logger=self.log, name="vt")

    def _validate_delay(self, delay):
        ''' Validate delay  parameter '''
        return validate_int(delay, const.DELAY_MIN, const.DELAY_MAX, logger=self.log, name="delay")

    def _validate_maxsize(self, maxsize):
        ''' Validate maxsize parameter '''
        return maxsize == -1 or validate_int(maxsize, const.MAXSIZE_MIN, const.MAXSIZE_MAX,
                                             logger=self.log, name="maxsize")

    def _set_param(self, name, value):
        ''' set parameter, with validation '''
        if name in self.PARAMS:
            if self._validate_param(name, value):
                self._params[name] = value
                return True
        return False

    def ready(self):
        ''' Check if we are ready to execute '''
        for param in self._required_params():
            if not self._validate_param(param, self._params[param]):
                self.log.info("Invalid value for parameter '%s'", param)
                return False
        return True

    def execute(self):
        ''' Execute Command '''
        if self._exceptions:
            return self._exec()
        try:
            ret = self._exec()
        except RedisSMQException as ex:
            if self.get_quiet is not True:
                self.log.warning("%s: Exception while processing %s: %s", ex.__class__.__name__,
                                 self.__class__.__name__, ex)
            return False
        return ret

    def _exec(self):
        if self.ready():
            return self.exec_command()
        return False

    def exec_command(self):
        ''' Execute Command '''
        raise CommandNotImplementedException()

    def queue_def(self):
        ''' Get Queue Definition '''
        queue_key = self.queue_key
        tx = self.client.pipeline(transaction=True)
        tx.hmget(queue_key, "vt", "delay", "maxsize")
        tx.time()

        results = tx.execute()

        if not results or results[0][0] is None:
            raise QueueDoesNotExist(self.get_queue)
        stats = results[0]
        redisTime = results[1]

        ts_usec = redisTime[0] * 1000000 + redisTime[1]
        ts = int(ts_usec / 1000)

        return {
            'qname': self.get_qname,
            "vt": stats[0],
            "delay": stats[1],
            "maxsize": stats[2],
            'ts': ts,
            'ts_usec': ts_usec
        }



'''

'''

from .base_command import BaseRSMQCommand
from .exceptions import NoMessageInQueue


class ReceiveMessageCommand(BaseRSMQCommand):
    '''
    Receive Message
    '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'vt': {'required': False,
                     'value': None},
              'quiet': {'required': False,
                        'value': False}
              }

    def exec_command(self):
        ''' Execute '''
        client = self.client

        queue_base = self.queue_base
        queue = self.queue_def()

        ts = int(queue['ts'])
        vt = int(queue['vt'] if self.get_vt is None else self.get_vt)
        vtimeout = ts + (vt * 1000)
        result = client.evalsha(
            self.receiveMessageSha1, 3, queue_base, ts, vtimeout)
        if not result:
            raise NoMessageInQueue(self.get_qname)
        [message_id, message, rc, ts] = result
        return {'id': message_id, 'message': message, 'rc': rc, 'ts': ts}



'''

'''
import time

from .. import const
from .base_command import BaseRSMQCommand
from .exceptions import QueueAlreadyExists


class CreateQueueCommand(BaseRSMQCommand):
    '''
    Create Queue if does not exist
    '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'vt': {'required': True,
                     'value': const.VT_DEFAULT},
              'delay': {'required': True,
                        'value': const.DELAY_DEFAULT},
              'maxsize': {'required': True,
                          'value': const.MAXSIZE_DEFAULT},
              'quiet': {'required': False,
                        'value': False}
              }

    def exec_command(self):
        ''' Exec Command '''
        client = self.client
        now = int(time.time())

        key = self.queue_key
        tx = client.pipeline(transaction=True)
        tx.hsetnx(key, "vt", self.get_vt)
        tx.hsetnx(key, "delay", self.get_delay)
        tx.hsetnx(key, "maxsize", self.get_maxsize)
        tx.hsetnx(key, "created", now)
        tx.hsetnx(key, "modified", now)
        results = tx.execute()
        if True not in results:
            raise QueueAlreadyExists(self.get_qname)
        client.sadd(self.queue_set, self.get_qname)
        return True



''' Utilities '''
import json
import random

DEFAULT_CHARSET = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"


def validate_int(value, min_value=None, max_value=None, logger=None, name=None):
    ''' Validate value is integer and between min and max values (if specified) '''
    if value is None:
        if logger and name:
            logger.debug("%s value is not set", name)
        return False
    try:
        int_value = int(value)
        if min_value is not None and int_value < min_value:
            if logger and name:
                logger.debug(
                    "%s value %s is less than minimum (%s)", name, int_value, min_value)
            return False
        if max_value is not None and int_value > max_value:
            if logger and name:
                logger.debug(
                    "%s value %s is greater than maximum (%s)", name, int_value, max_value)
            return False
    except ValueError:
        if logger and name:
            logger.debug("%s value (%s) is not an integer", name, value)
        return False
    return True


def baseXencode(value, chars='0123456789abcdefghijklmnopqrstuvwxyz'):
    '''
    Converts an integer to a base X string using charset. 
    Base is implied by charset = base36 by default 

    Based on: https://en.wikipedia.org/wiki/Base36

    raises ValueError if value cannot be converted to integer
    '''

    base = len(chars)
    integer = int(value)
    sign = '-' if integer < 0 else ''
    integer = abs(integer)
    result = ''

    while integer > 0:
        integer, remainder = divmod(integer, base)
        result = chars[remainder] + result

    return sign + result


def random_string(length, charset=None):
    ''' generate a random string of characters from charset '''
    if not charset:
        charset = DEFAULT_CHARSET

    string = ''
    for _ in range(length):
        string += random.choice(charset)
    return string


def make_message_id(usec):
    ''' Create a message id based on Redis time '''
    return baseXencode(usec) + random_string(22)


def encode_message(msg):
    ''' Encode message to JSON if not already string '''
    if isinstance(msg, str):
        return msg
    return json.dumps(msg)


def decode_message(msg):
    ''' Decode message from JSON if decodable, else return message as is '''
    if isinstance(msg, str):
        try:
            return json.loads(msg)
        except json.decoder.JSONDecodeError:
            pass
    return msg



'''
Example Consumer for a queue


In this example we create an queue consumer that gets one message at a time, and processes
it using a processor. To emulate some of the complexities, we randomly fail processing and adding
delays to processing.

'''
import argparse
import logging
import random
import sys
from threading import Thread
import time

from rsmq import RedisSMQ
from rsmq.consumer import RedisSMQConsumer


LOG = logging.getLogger("Consumer")


class Processor(object):
    ''' Dummy processor class to fake complexity of a real processor '''

    def __init__(self, delay, success_rate):
        self.delay = delay
        self.success_rate = success_rate

    def random_result(self):
        ''' 
        Produce a random boolean which is True "success_rate" percent of the time

        random.random() produces a float in range [0.0, 1.0)
        '''
        return random.random() < self.success_rate

    def process(self, id, message, rc, ts):
        ''' Actual method that processes the message '''
        LOG.info("Got message: id: %s, retry count: %s, ts: %s, msg: %s" %
                 (id, rc, ts, message))
        result = self.random_result()
        delay = self.delay
        if delay > 0:
            # Add occasional long delay
            # if random.randint(0, 10) == 0:
            #    delay = 60
            LOG.info("Processing message: %s for %s seconds", id, delay)
            time.sleep(delay)
        #LOG.info("Random Result: %s", "Success" if result else "Failure")
        return result


def consume(consumer, long_qname, exit_after):
    ''' Example of consuming using a thread '''
    LOG.info("Starting consumption on queue: %s", long_qname)
    thread = Thread(
        target=consumer.run, name="QueueConumer", daemon=False)
    thread.start()
    end = time.time() + exit_after if exit_after else 0
    while True:
        if exit_after and time.time() > end:
            LOG.info("Attempting to stop the consumer...")
            consumer.stop(15)
            break
    LOG.info("Exited queue consumer for '%s'. Thread is %s",
             long_qname, ("running" if thread.is_alive() else "stopped"))


def probablity(x):
    ''' An arg validator that Require a float number between 0.0 and 1.0 '''
    x = float(x)
    if not (0.0 <= x <= 1.0):
        raise argparse.ArgumentTypeError("%r not in range [0.0, 1.0]" % (x,))
    return x


def main(argv=None):
    if argv is None:
        argv = sys.argv
    ''' Parse args and run producer '''
    parser = argparse.ArgumentParser()
    parser.add_argument("-q", "--queue", dest="queue", action="store", default="queue",
                        help="queue name [default: %(default)s]")
    parser.add_argument("-n", "--namespace", dest="ns", action="store", default="test",
                        help="queue namespace [default: %(default)s]")

    parser.add_argument("-r", "--success_rate", dest="success_rate", action="store",
                        type=probablity,
                        default=1.0, help="Probability of success when processing messages." +
                        "1.0 means all are successful, 0.0 means all fail" +
                        "[default: %(default)s]")

    parser.add_argument("-e", "--empty_delay", dest="empty_queue_delay", type=float, default=4.0,
                        help="delay in seconds when queue is empty[default: %(default)s]")

    parser.add_argument("-d", "--delay", dest="delay", type=float, default=2.0,
                        help="additional delay in seconds, during consumption[default: %(default)s]")

    parser.add_argument("-x", "--exit_after", dest="exit_after", type=float, default=0.0,
                        help="If set, exit after this many seconds[default: %(default)s]")

    parser.add_argument("-v", "--visibility_timeout", dest="vt", type=int, default=None,
                        help="Visibility Timeout[default: %(default)s]")

    parser.add_argument("--no-trace", dest="trace", action="store_false", default=True,
                        help="If set, hide trace messages")

    parser.add_argument("-H", dest="host", default="127.0.0.1",
                        help="Redis Host [default: %(default)s]")
    parser.add_argument("-P", dest="port", type=int, default=6379,
                        help="Redis Port [default: %(default)s]")

    # Parse command line args`
    args = parser.parse_args()

    # Create Processor
    processor = Processor(delay=args.delay, success_rate=args.success_rate)

    # Create RedisSMQ queue consumer controller
    LOG.info("Creating RedisSMQ Consumer Controller for redis at %s:%s, using queue: %s:%s",
             args.host, args.port, args.ns, args.queue)
    rsqm_consumer = RedisSMQConsumer(qname=args.queue,
                                     processor=processor.process,
                                     host=args.host, port=args.port, ns=args.ns, vt=args.vt,
                                     empty_queue_delay=args.empty_queue_delay, trace=args.trace)

    # Start Consumption
    consume(rsqm_consumer, "%s:%s" % (args.ns, args.queue), args.exit_after)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    main(sys.argv)



'''
Baseline Unit Test
'''
import difflib
import os.path
import pprint
import unittest
from unittest.util import safe_repr

import fakeredis

from rsmq.rsmq import RedisSMQ


BASE_DIR = os.path.dirname(os.path.dirname(__file__))


class TestUnitTests(unittest.TestCase):
    ''' Unit Test RedisSMQ'''

    def setUp(self):
        ''' Setup '''

    def assertQueueAttributes(self, expected, actual, msg=None):
        ''' assert queue attributes are set '''
        self.assertIsInstance(
            expected, dict, 'First argument is not a dictionary')
        self.assertIsInstance(
            actual, dict, 'Second argument is not a dictionary')
        actual_short = {key: actual[key] for key in expected.keys()}
        if not expected.items() <= actual.items():
            diff = ('\n' + '\n'.join(difflib.ndiff(
                           pprint.pformat(expected).splitlines(),
                           pprint.pformat(actual_short).splitlines())))
            standardMsg = '%s !<= %s' % (
                safe_repr(expected), safe_repr(actual_short))
            standardMsg = self._truncateMessage(standardMsg, diff)
            msg = self._formatMessage(msg, standardMsg)
            raise AssertionError(msg)

    def test_unittest(self):
        ''' Check Unit Test Framework '''
        unittest_loaded = True
        self.assertTrue(unittest_loaded)

    def test_qname_good_validation(self):
        ''' Test Good QName validation '''
        client = fakeredis.FakeStrictRedis(decode_responses=True)
        client.flushall()
        queue = RedisSMQ(client=client, exceptions=False)
        GOOD_QUEUE_NAMES = ['simple', 'with_underscore', 'with-dash', '-start',
                            'with\x01\x02\x03binary']

        for qname in GOOD_QUEUE_NAMES:
            self.assertTrue(queue.getQueueAttributes().qname(qname).ready())

    def test_qname_bad_validation(self):
        ''' Test bad QName validation '''
        client = fakeredis.FakeStrictRedis(decode_responses=True)
        client.flushall()
        queue = RedisSMQ(client=client, exceptions=False)
        BAD_QUEUE_NAMES = ['', None, 'something:else',
                           ':', 'something:', ':else']

        for qname in BAD_QUEUE_NAMES:
            self.assertFalse(queue.getQueueAttributes().qname(qname).ready())

    def test_create_queue_default(self):
        ''' Test Creating of the Queue '''
        client = fakeredis.FakeStrictRedis(decode_responses=True)
        client.flushall()
        queue = RedisSMQ(client=client)
        queue_name = 'test-queue'
        queue_key = 'rsmq:%s:Q' % queue_name
        queue.createQueue().qname(queue_name).execute()
        keys = client.keys('*')
        self.assertListEqual(sorted([queue_key, 'rsmq:QUEUES']), sorted(keys))
        queues = client.smembers('rsmq:QUEUES')
        self.assertSetEqual(set([queue_name]), queues)
        queue_details = client.hgetall(queue_key)
        self.assertTrue(
            {'vt': '30', 'delay': '0', 'maxsize': '65565'}.items() <= queue_details.items())

    def test_create_queue_custom(self):
        ''' Test Creating of the Queue '''
        client = fakeredis.FakeStrictRedis(decode_responses=True)
        client.flushall()
        queue = RedisSMQ(client=client)
        queue_name = 'test-queue-2'
        queue_key = 'rsmq:%s:Q' % queue_name
        queue.createQueue(qname=queue_name, vt=15).delay(10).execute()
        keys = client.keys('*')
        self.assertListEqual(sorted([queue_key, 'rsmq:QUEUES']), sorted(keys))
        queues = client.smembers('rsmq:QUEUES')
        self.assertSetEqual(set([queue_name]), queues)
        queue_details = client.hgetall(queue_key)

        self.assertTrue(
            {'vt': '15', 'delay': '10', 'maxsize': '65565'}.items() <= queue_details.items())

    def test_delete_queue(self):
        ''' Test Deleting of the Queue '''
        client = fakeredis.FakeStrictRedis(decode_responses=True)
        client.flushall()
        queue = RedisSMQ(client=client)
        queue_name = 'test-queue-delete'
        queue_key = 'rsmq:%s:Q' % queue_name
        queue.createQueue().qname(queue_name).execute()
        keys = client.keys('*')
        self.assertListEqual(sorted([queue_key, 'rsmq:QUEUES']), sorted(keys))
        queues = client.smembers('rsmq:QUEUES')
        self.assertSetEqual(set([queue_name]), queues)
        queue_details = client.hgetall(queue_key)
        self.assertTrue(
            {'vt': '30', 'delay': '0', 'maxsize': '65565'}.items() <= queue_details.items())
        result = queue.deleteQueue(qname=queue_name).execute()
        self.assertEqual(result, True)
        queue_details = client.hgetall(queue_key)
        self.assertEqual(queue_details, {})

    def test_queue_attributes(self):
        ''' Test Getting/Setting of the Queue Attribues'''
        client = fakeredis.FakeStrictRedis(decode_responses=True)
        client.flushall()
        queue_name = 'test-queue-attr'
        queue = RedisSMQ(client=client, qname=queue_name)
        queue_key = 'rsmq:%s:Q' % queue_name
        queue.createQueue(vt=15, delay=10).execute()
        keys = client.keys('*')
        self.assertListEqual(sorted([queue_key, 'rsmq:QUEUES']), sorted(keys))
        queues = client.smembers('rsmq:QUEUES')
        self.assertSetEqual(set([queue_name]), queues)
        queue_details = client.hgetall(queue_key)
        self.assertQueueAttributes(
            {'vt': '15', 'delay': '10', 'maxsize': '65565'}, queue_details)
        attributes = queue.getQueueAttributes(qname=queue_name).execute()
        self.assertQueueAttributes({'vt': '15',
                                    'delay': '10',
                                    'maxsize': '65565',
                                    'totalrecv': 0,
                                    'totalsent': 0,
                                    'msgs': 0,
                                    'hiddenmsgs': 0}, attributes)
        attributes_1 = queue.setQueueAttributes(
            vt=30, maxsize=1024, delay=1).execute()
        attributes = queue.getQueueAttributes(qname=queue_name).execute()
        self.assertDictEqual(attributes, attributes_1)
        self.assertQueueAttributes({'vt': '30',
                                    'delay': '1',
                                    'maxsize': '1024',
                                    'totalrecv': 0,
                                    'totalsent': 0,
                                    'msgs': 0,
                                    'hiddenmsgs': 0}, attributes)

    def test_invalid_queue_attributes(self):
        ''' Test Getting/Setting of the Queue Attribues'''
        client = fakeredis.FakeStrictRedis(decode_responses=True)
        client.flushall()
        queue_name = 'test-queue-attr'
        queue = RedisSMQ(client=client, qname=queue_name)
        queue_key = 'rsmq:%s:Q' % queue_name
        queue.createQueue(vt=15, delay=10).execute()
        keys = client.keys('*')
        self.assertListEqual(sorted([queue_key, 'rsmq:QUEUES']), sorted(keys))
        queues = client.smembers('rsmq:QUEUES')
        self.assertSetEqual(set([queue_name]), queues)
        queue_details = client.hgetall(queue_key)
        self.assertQueueAttributes(
            {'vt': '15', 'delay': '10', 'maxsize': '65565'}, queue_details)
        attributes = queue.getQueueAttributes(qname=queue_name).execute()
        self.assertQueueAttributes({'vt': '15',
                                    'delay': '10',
                                    'maxsize': '65565',
                                    'totalrecv': 0,
                                    'totalsent': 0,
                                    'msgs': 0,
                                    'hiddenmsgs': 0}, attributes)
        attributes_1 = queue.setQueueAttributes(
            vt=-3, maxsize=-2, delay=-1).execute()
        attributes = queue.getQueueAttributes(qname=queue_name).execute()
        self.assertDictEqual(attributes, attributes_1)
        self.assertQueueAttributes({'vt': '15',
                                    'delay': '10',
                                    'maxsize': '65565',
                                    'totalrecv': 0,
                                    'totalsent': 0,
                                    'msgs': 0,
                                    'hiddenmsgs': 0}, attributes)


if __name__ == '__main__':
    unittest.main()



'''
Example Consumer for a queue


In this example we create an queue consumer that gets one message at a time, and processes
it using a processor. To emulate some of the complexities, we randomly fail processing and adding
delays to processing.

This example is similar to consumer.py but uses RedisSMQConsumerThread which is a versioin 
of RedisSMQConsumer that extends the Thread class
'''
import argparse
import logging
import random
import sys
import time

from rsmq import RedisSMQ
from rsmq.consumer import RedisSMQConsumerThread


LOG = logging.getLogger("Consumer")


class Processor(object):
    ''' Dummy processor class to fake complexity of a real processor '''

    def __init__(self, delay, success_rate):
        self.delay = delay
        self.success_rate = success_rate

    def random_result(self):
        ''' 
        Produce a random boolean which is True "success_rate" percent of the time

        random.random() produces a float in range [0.0, 1.0)
        '''
        return random.random() < self.success_rate

    def process(self, id, message, rc, ts):
        ''' Actual method that processes the message '''
        LOG.info("Got message: id: %s, retry count: %s, ts: %s, msg: %s (%s)" %
                 (id, rc, ts, message, type(message)))
        result = self.random_result()
        delay = self.delay
        if delay > 0:
            # Add occasional long delay
            # if random.randint(0, 10) == 0:
            #    delay = 60
            LOG.info("Processing message: %s for %s seconds", id, delay)
            time.sleep(delay)
        #LOG.info("Random Result: %s", "Success" if result else "Failure")
        return result


def consume(consumer, long_qname, exit_after):
    ''' Example of consuming using a thread '''
    LOG.info("Starting consumption on queue: %s", long_qname)
    consumer.start()
    end = time.time() + exit_after if exit_after else 0
    while True:
        if exit_after and time.time() > end:
            LOG.info("Attempting to stop the consumer...")
            consumer.stop(15)
            break
    LOG.info("Exited queue consumer for '%s'. Thread is %s",
             long_qname, ("running" if consumer.is_alive() else "stopped"))


def probablity(x):
    ''' An arg validator that Require a float number between 0.0 and 1.0 '''
    x = float(x)
    if not (0.0 <= x <= 1.0):
        raise argparse.ArgumentTypeError("%r not in range [0.0, 1.0]" % (x,))
    return x


def main(argv=None):
    if argv is None:
        argv = sys.argv
    ''' Parse args and run producer '''
    parser = argparse.ArgumentParser()
    parser.add_argument("-q", "--queue", dest="queue", action="store", default="queue",
                        help="queue name [default: %(default)s]")
    parser.add_argument("-n", "--namespace", dest="ns", action="store", default="test",
                        help="queue namespace [default: %(default)s]")

    parser.add_argument("-r", "--success_rate", dest="success_rate", action="store",
                        type=probablity,
                        default=1.0, help="Probability of success when processing messages." +
                        "1.0 means all are successful, 0.0 means all fail" +
                        "[default: %(default)s]")

    parser.add_argument("-e", "--empty_delay", dest="empty_queue_delay", type=float, default=4.0,
                        help="delay in seconds when queue is empty[default: %(default)s]")

    parser.add_argument("-d", "--delay", dest="delay", type=float, default=2.0,
                        help="additional delay in seconds, during consumption[default: %(default)s]")

    parser.add_argument("-x", "--exit_after", dest="exit_after", type=float, default=0.0,
                        help="If set, exit after this many seconds[default: %(default)s]")

    parser.add_argument("-v", "--visibility_timeout", dest="vt", type=int, default=None,
                        help="Visibility Timeout[default: %(default)s]")

    parser.add_argument("--no-trace", dest="trace", action="store_false", default=True,
                        help="If set, hide trace messages")

    parser.add_argument("-D", "--decode", dest="decode", action="store_true", default=False,
                        help="If set, decode messages from JSON")

    parser.add_argument("-H", dest="host", default="127.0.0.1",
                        help="Redis Host [default: %(default)s]")
    parser.add_argument("-P", dest="port", type=int, default=6379,
                        help="Redis Port [default: %(default)s]")

    # Parse command line args`
    args = parser.parse_args()

    # Create Processor
    processor = Processor(delay=args.delay, success_rate=args.success_rate)

    # Create RedisSMQ queue consumer controller
    LOG.info("Creating RedisSMQ Consumer Controller for redis at %s:%s, using queue: %s:%s",
             args.host, args.port, args.ns, args.queue)
    rsqm_consumer = RedisSMQConsumerThread(qname=args.queue,
                                           processor=processor.process,
                                           host=args.host, port=args.port, ns=args.ns, vt=args.vt,
                                           empty_queue_delay=args.empty_queue_delay,
                                           decode=args.decode, trace=args.trace)

    # Start Consumption
    consume(rsqm_consumer, "%s:%s" % (args.ns, args.queue), args.exit_after)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    main(sys.argv)



'''

'''

from .base_command import BaseRSMQCommand


class DeleteMessageCommand(BaseRSMQCommand):
    '''
    Delete Message if it exists
    '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'id': {'required': True,
                     'value': None},
              'quiet': {'required': False,
                        'value': False}
              }

    def exec_command(self):
        ''' Exec Command '''
        client = self.client
        queue_base = self.queue_base
        queue_key = self.queue_key
        message_id = self.get_id

        tx = client.pipeline(transaction=True)
        tx.zrem(queue_base, message_id)
        tx.hdel(queue_key, message_id, "%s:rc" %
                message_id, "%s:fr" % message_id)
        result = tx.execute()

        if int(result[0]) == 1 and int(result[1]) > 0:
            return True

        return False



'''
Unit Tests for utils
'''
import logging
import os
import unittest

from rsmq.cmd import utils


BASE_DIR = os.path.dirname(os.path.dirname(__file__))


ENCODE_TESTS = {
    'a': 'a',
    '{"a": "b"}': {'a': 'b'}
}

# Each test is value, min, max, sucess
VALIDATE_INT_TESTS = [
    (0, 0, 100, True),
    (1, 0, 100, True),
    (1, 0, 1, True),
    (-1, -100, 0, True),
    (0, -100, 0, True),
    (1, -100, 0, False),
    (1, 0, 1, True),
    (0, 1, 2, False),
]

# Base charsets
BASE_10 = '0123456789'
BASE_16 = '0123456789ABCDEF'
BASE_36 = '0123456789abcdefghijklmnopqrstuvwxyz'

# Each test is (raw, encoded, base_charset)
BASE_X_ENCODED_TESTS = [
    (10000, "10000", BASE_10),
    (10000, "2710", BASE_16),
    (10000, "7ps", BASE_36),
    (10000000, "5yc1s", BASE_36),
]


class MockLogger(object):

    def __init__(self):
        self.last_level = None
        self.last_msg = None

    def reset(self):
        self.last_level = None
        self.last_msg = None

    def log(self, level, msg, *params):
        self.last_level = level
        self.last_msg = msg % params

    def debug(self, msg, *params):
        self.log(logging.DEBUG, msg, *params)

    def info(self, msg, *params):
        self.log(logging.INFO, msg, *params)


class CmdUtilsUnitTests(unittest.TestCase):
    ''' Unit Tests for utils'''

    def setUp(self):
        ''' Setup '''

    def test_encode_object(self):
        ''' Check Encode Message '''
        for string, obj in ENCODE_TESTS.items():
            self.assertEqual(string, utils.encode_message(obj))

    def test_encode_string(self):
        ''' Check Encode String Message '''
        for string, _obj in ENCODE_TESTS.items():
            self.assertEqual(string, utils.encode_message(string))

    def test_decode_object(self):
        ''' Check Decode Message Object '''
        for _string, obj in ENCODE_TESTS.items():
            self.assertEqual(obj, utils.decode_message(obj))

    def test_decode_string(self):
        ''' Check Decode Message String '''
        for string, obj in ENCODE_TESTS.items():
            self.assertEqual(obj, utils.decode_message(string))

    def test_validate_int(self):
        ''' Test validate_int '''
        for value, min_value, max_value, expected in VALIDATE_INT_TESTS:
            self.assertEqual(expected, utils.validate_int(
                value, min_value, max_value))

    def test_validate_int_invalid(self):
        ''' Test validate_int '''
        self.assertFalse(utils.validate_int(None, 0, 100))
        self.assertFalse(utils.validate_int("A", 0, 100))

    def test_validate_int_logging_none(self):
        ''' Test validate_int '''
        logger = MockLogger()
        self.assertFalse(utils.validate_int(None, 0, 100, logger, "testname"))
        self.assertEqual("testname value is not set", logger.last_msg)

    def test_validate_int_logging_too_low(self):
        ''' Test validate_int '''
        logger = MockLogger()
        self.assertFalse(utils.validate_int(-1, 0, 100, logger, "testname"))
        self.assertEqual(
            "testname value -1 is less than minimum (0)", logger.last_msg)

    def test_validate_int_logging_too_high(self):
        ''' Test validate_int '''
        logger = MockLogger()
        self.assertFalse(utils.validate_int(101, 0, 100, logger, "testname"))
        self.assertEqual("testname value 101 is greater than maximum (100)",
                         logger.last_msg)

    def test_validate_int_logging_invalid(self):
        ''' Test validate_int '''
        logger = MockLogger()
        self.assertFalse(utils.validate_int("A", 0, 100, logger, "testname"))
        self.assertEqual(
            "testname value (A) is not an integer", logger.last_msg)

    def test_random_string(self):
        ''' Test random_string '''
        charset_default = set(utils.DEFAULT_CHARSET)
        charset1 = "abcdefgh"
        charset1_set = set(charset1)

        for i in range(32):
            for _ in range(10):
                rstr = utils.random_string(i, charset1)
                self.assertEqual(
                    i, len(rstr), "Expected length is %s, got %s" % (i, len(rstr)))
                self.assertLessEqual(
                    set(rstr), charset1_set, "Expected all characters to be in set")
            for _ in range(10):
                rstr = utils.random_string(i)
                self.assertEqual(
                    i, len(rstr), "Expected length is %s, got %s" % (i, len(rstr)))
                self.assertLessEqual(
                    set(rstr), charset_default, "Expected all characters to be in set")

    def test_make_message_id(self):
        ''' Test make_message_id '''
        for _ in range(32):
            mid = utils.make_message_id(10000000)
            # length is always 22 plus encoded value of 10000000 = which is
            self.assertEqual(27, len(mid))

    def test_baseXencode(self):
        ''' Test baseXencode '''
        for (raw, encoded, charset) in BASE_X_ENCODED_TESTS:
            self.assertEqual(encoded, utils.baseXencode(
                raw, charset), "Invalid encoded value for base %s" % len(charset))


if __name__ == '__main__':
    unittest.main()



'''
Python Redis Simple Queue Manager
'''

from redis import Redis

from . import const
from .cmd import ChangeMessageVisibilityCommand
from .cmd import CreateQueueCommand, DeleteQueueCommand, ListQueuesCommand
from .cmd import DeleteMessageCommand
from .cmd import SendMessageCommand, ReceiveMessageCommand, PopMessageCommand
from .cmd import SetQueueAttributesCommand, GetQueueAttributesCommand


DEFAULT_REDIS_OPTIONS = {
    'encoding': 'utf-8',
    'decode_responses': True
}

DEFAULT_OPTIONS = {
    'ns': 'rsmq',
    'realtime': False,
    'exceptions': True
}


class RedisSMQ():
    '''
    Redis Simple Message Queue implementation in Python
    '''

    def __init__(self, client=None, host="127.0.0.1", port="6379", options=None, **kwargs):
        '''
        Constructor:

        @param client: if provided, redis client object to use
        @param host: if client is not provided, redis hostname
        @param port: if client is not provided, redis port
        @param options: if client is not provided, additional options for redis client creation

        Additional arguments:

        @param ns: namespace
        @param realtime: if true, use realtime comms (pubsub)(default is False)
        @param exceptions: if true, throw exceptions on errors, else return False(default is True)

        Remaining params are automatically passed to commands

        '''

        # redis_client
        self._client = client

        # Redis Options
        self.redis_options = dict(DEFAULT_REDIS_OPTIONS)
        self.redis_options['host'] = host
        self.redis_options['port'] = port
        if options:
            self.redis_options.update(options)

        # RSMQ global options
        self.options = dict(DEFAULT_OPTIONS)
        to_remove = []
        for param, value in kwargs.items():
            if param in self.options:
                self.options[param] = kwargs[param]
                to_remove.append(param)
            elif value is None:
                # Remove default set to None
                to_remove.append(param)

        # Remove unnecessary kwargs
        for param in to_remove:
            del kwargs[param]

        # Everything else is passed through to commands
        self._default_params = kwargs

        self._popMessageSha1 = None
        self._receiveMessageSha1 = None
        self._changeMessageVisibilitySha1 = None

    @property
    def popMessageSha1(self):
        ''' Get Pop Message Script SHA1 '''
        if self._popMessageSha1 is None:
            client = self.client
            self._popMessageSha1 = client.script_load(const.SCRIPT_POPMESSAGE)
        return self._popMessageSha1

    @property
    def receiveMessageSha1(self):
        ''' Get Received Message Script SHA1 '''
        if self._receiveMessageSha1 is None:
            client = self.client
            self._receiveMessageSha1 = client.script_load(
                const.SCRIPT_RECEIVEMESSAGE)
        return self._receiveMessageSha1

    @property
    def changeMessageVisibilitySha1(self):
        ''' Get Change Message Visibilities Script SHA1 '''
        if self._changeMessageVisibilitySha1 is None:
            client = self.client
            self._changeMessageVisibilitySha1 = client.script_load(
                const.SCRIPT_CHANGEMESSAGEVISIBILITY)
        return self._changeMessageVisibilitySha1

    @property
    def client(self):
        ''' get Redis client. Create one if one does not exist '''
        if not self._client:
            self._client = Redis(**self.redis_options)
        return self._client

    def exceptions(self, enabled=True):
        ''' Set global exceptions flag '''
        self.options['exceptions'] = enabled == True
        return self

    def setCient(self, client):
        ''' Set Redis Client '''
        self._client = client
        return self

    def _command(self, command, **kwargs):
        ''' Run command '''
        args = dict(self._default_params)
        args.update(kwargs)
        return command(self, **args)

    def createQueue(self, **kwargs):
        ''' Create Queue '''
        return self._command(CreateQueueCommand, **kwargs)

    def deleteQueue(self, **kwargs):
        ''' Create Queue '''
        return self._command(DeleteQueueCommand, **kwargs)

    def setQueueAttributes(self, **kwargs):
        ''' setQueueAttributesCommand() '''
        return self._command(SetQueueAttributesCommand, **kwargs)

    def getQueueAttributes(self, **kwargs):
        ''' getQueueAttributesCommand() '''
        return self._command(GetQueueAttributesCommand, **kwargs)

    def listQueues(self, **kwargs):
        ''' List Queues '''
        return self._command(ListQueuesCommand, **kwargs)

    def changeMessageVisibility(self, **kwargs):
        ''' ChangeMessageVisibilityCommand '''
        return self._command(ChangeMessageVisibilityCommand, **kwargs)

    def sendMessage(self, **kwargs):
        ''' Send Message Command '''
        return self._command(SendMessageCommand, **kwargs)

    def receiveMessage(self, **kwargs):
        ''' Receive Message Command '''
        return self._command(ReceiveMessageCommand, **kwargs)

    def popMessage(self, **kwargs):
        ''' Pop Message Command '''
        return self._command(PopMessageCommand, **kwargs)

    def deleteMessage(self, **kwargs):
        ''' Delete Message Command '''
        return self._command(DeleteMessageCommand, **kwargs)

    def quit(self):
        ''' Quit - here for compatibility purposes '''
        self._client = None




from pprint import pprint
import time

from rsmq.rsmq import RedisSMQ


# Create controller.
# In this case we are specifying the host and default queue name
queue = RedisSMQ(host="127.0.0.1", qname="myqueue")


# Delete Queue if it already exists, ignoring exceptions
queue.deleteQueue().exceptions(False).execute()

# Create Queue with default visibility timeout of 20 and delay of 0
# demonstrating here both ways of setting parameters
queue.createQueue(delay=0).vt(20).execute()

# Send a message with a 2 second delay
message_id = queue.sendMessage(delay=2).message("Hello World").execute()

pprint({'queue_status': queue.getQueueAttributes().execute()})

# Try to get a message - this will not succeed, as our message has a delay and no other
# messages are in the queue
msg = queue.receiveMessage().exceptions(False).execute()

# Message should be False as we got no message
pprint({"Message": msg})

print("Waiting for our message to become visible")
# Wait for our message to become visible
time.sleep(2)

pprint({'queue_status': queue.getQueueAttributes().execute()})
# Get our message
msg = queue.receiveMessage().execute()

# Message should now be there
pprint({"Message": msg})

# Delete Message
queue.deleteMessage(id=msg['id'])

pprint({'queue_status': queue.getQueueAttributes().execute()})
# delete our queue
queue.deleteQueue().execute()

# No action
queue.quit()



'''
Delete Queue Command
'''

from .base_command import BaseRSMQCommand
from .exceptions import QueueDoesNotExist


class DeleteQueueCommand(BaseRSMQCommand):
    '''
    Delete Queue if it exists
    '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'quiet': {'required': False,
                        'value': False}
              }

    def exec_command(self):
        ''' Exec Command '''
        client = self.client

        queue_name = self.queue_base
        queue_key = self.queue_key
        tx = client.pipeline(transaction=True)
        tx.delete(queue_key)
        tx.delete(queue_name)
        tx.srem(self.queue_set, self.get_qname)
        results = tx.execute()
        if True not in results:
            raise QueueDoesNotExist(self.get_qname)
        self.log.debug("Deleted Queue %s", self.queue_base)
        return True



'''
Command Exceptions 
'''


class RedisSMQException(Exception):
    ''' Base Exception for all RQSM Commands '''

    def __init__(self, msg=None):
        ''' Default constructor '''
        if msg is None:
            msg = self.__doc__
        super(RedisSMQException, self).__init__(msg)


class CommandNotImplementedException(RedisSMQException):
    ''' Command is not yet implemented '''


class InvalidParameterValue(RedisSMQException):
    ''' Invalid Parameter Value'''

    def __init__(self, name, value):
        ''' Constructor '''
        super(InvalidParameterValue, self).__init__(
            "Value '%s' is not valid for parameter '%s'" % (value, name))


class QueueAlreadyExists(RedisSMQException):
    ''' Queue Already Exists '''

    def __init__(self, name):
        ''' Constructor '''
        super(QueueAlreadyExists, self).__init__(
            "Queue '%s' already exists" % name)


class QueueDoesNotExist(RedisSMQException):
    ''' Queue Already Exists '''

    def __init__(self, name):
        ''' Constructor '''
        super(QueueDoesNotExist, self).__init__(
            "Queue '%s' does not exist" % name)


class NoMessageInQueue(RedisSMQException):
    ''' Queue Already Exists '''

    def __init__(self, name):
        ''' Constructor '''
        super(NoMessageInQueue, self).__init__(
            "Queue '%s' has no messages waiting" % name)



''' 
Python Implementation of the Redis Simple Message Queue  serverless(*) queue service

(* requires only Redis server)

Based on:

  * Original Node.js version: https://github.com/smrchy/rsmq
  * Java port: https://github.com/igr/jrsmq

'''

from .rsmq import RedisSMQ



'''

'''

from .base_command import BaseRSMQCommand
from .utils import make_message_id, encode_message


class SendMessageCommand(BaseRSMQCommand):
    '''
    Create Queue if does not exist
    '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'message': {'required': True,
                          'value': None},
              'delay': {'required': False,
                        'value': None},
              'quiet': {'required': False,
                        'value': False},
              'encode': {'required': False,
                         'value': False}
              }

    def exec_command(self):
        ''' 
        Execute command

        @raise QueueDoesNotExist if queue does not exist
        '''
        queue = self.queue_def()
        message_id = make_message_id(queue.get('ts_usec', None))

        queue_key = self.queue_key
        queue_base = self.queue_base

        ts = int(queue['ts'])

        delay = self.get_delay
        if delay is None:
            delay = queue.get('delay', 0)
        delay = int(delay or 0)

        message = self.get_message
        if self.get_encode or not isinstance(message, str):
            message = encode_message(message)
            self.log.debug("Encoded message: %s", message)

        tx = self.client.pipeline(transaction=True)
        timestamp = ts + delay * 1000
        self.log.debug("tx.zadd(%s, %s, %s)",
                       queue_base, timestamp, message_id)
        tx.zadd(queue_base, {message_id: timestamp})

        tx.hset(queue_key, message_id, message)
        tx.hincrby(queue_key, "totalsent", 1)
        _results = tx.execute()

        return message_id



'''
Example Producer for a queue


In this example we create RedisSMQ controller and using it to send messages to that at
regular intervals

'''
import argparse
import logging
import random
import sys
from threading import Thread
import time

from rsmq import RedisSMQ
from rsmq.consumer import RedisSMQConsumer


LOG = logging.getLogger("Producer")


def create_message(msg_number):
    ''' Create a message to send '''
    return {"item": msg_number,
            "ts": time.time()}


def produce(rsmq, long_qname, count, interval):
    LOG.info("Starting producer for queue '%s' - sending %s messages every %s seconds...",
             long_qname, count, interval)

    msg_count = 0
    while True:
        msg_count += 1
        msg = create_message(msg_count)
        rsmq.sendMessage(message=msg, encode=True).execute()
        if count > 0 and msg_count >= count:
            # Stop if we are done
            break
        LOG.debug("Waiting %s seconds to send next message", interval)
        time.sleep(interval)
    LOG.info("Ended producer after sending %s messages.", msg_count)


def main(argv=None):
    if argv is None:
        argv = sys.argv
    ''' Parse args and run producer '''
    parser = argparse.ArgumentParser()
    parser.add_argument("-q", "--queue", dest="queue", action="store", default="queue",
                        help="queue name [default: %(default)s]")
    parser.add_argument("-n", "--namespace", dest="ns", action="store", default="test",
                        help="queue namespace [default: %(default)s]")

    parser.add_argument("-c", "--count", dest="count", action="store", type=int, default=0,
                        help="number of messages to send. If less than 1, send continuously " +
                        "[default: %(default)s]")

    parser.add_argument("-i", "--interval", dest="interval", type=float, default=1.5,
                        help="Interval, in seconds, to send [default: %(default)s]")

    parser.add_argument("-d", "--delay", dest="delay", type=int, default=0,
                        help="delay, in seconds, to send message with [default: %(default)s]")

    parser.add_argument("-v", "--visibility_timeout", dest="vt", type=int, default=None,
                        help="Visibility Timeout[default: %(default)s]")

    parser.add_argument("--delete", dest="delete", action="store_true", default=False,
                        help="If set, delete queue first")

    parser.add_argument("--no-trace", dest="trace", action="store_false", default=True,
                        help="If set, hide trace messages")

    parser.add_argument("-H", dest="host", default="127.0.0.1",
                        help="Redis Host [default: %(default)s]")
    parser.add_argument("-P", dest="port", type=int, default=6379,
                        help="Redis Port [default: %(default)s]")
    # Parse command line args`
    args = parser.parse_args()

    # Create RedisSMQ queue controller
    LOG.info("Creating RedisSMQ controller for redis at %s:%s, using default queue: %s:%s",
             args.host, args.port, args.ns, args.queue)
    rsqm = RedisSMQ(qname=args.queue, host=args.host,
                    port=args.port, ns=args.ns, vt=args.vt,
                    delay=args.delay, trace=args.trace)

    if args.delete:
        rsqm.deleteQueue(qname=args.queue, quiet=True).exceptions(
            False).execute()

    # Create queue if it is missing. Swallow errors if already exists
    rsqm.createQueue(qname=args.queue, quiet=True).exceptions(False).execute()

    # Start Producing
    produce(rsqm, "%s:%s" % (args.ns, args.queue), args.count, args.interval)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    main(sys.argv)



'''
Commands
'''

from .change_message_visibility import ChangeMessageVisibilityCommand
from .create_queue import CreateQueueCommand
from .delete_message import DeleteMessageCommand
from .delete_queue import DeleteQueueCommand
from .exceptions import *
from .get_queue_attributes import GetQueueAttributesCommand
from .list_queues import ListQueuesCommand
from .pop_message import PopMessageCommand
from .receive_message import ReceiveMessageCommand
from .send_message import SendMessageCommand
from .set_queue_attributes import SetQueueAttributesCommand



'''
GetQueueAttributes Command
'''

import time

from .base_command import BaseRSMQCommand
from .exceptions import QueueDoesNotExist


class GetQueueAttributesCommand(BaseRSMQCommand):
    '''
    Get Queue Attributes from existing queue
    '''

    PARAMS = {'qname': {'required': True,
                        'value': None},
              'quiet': {'required': False,
                        'value': False}
              }

    def exec_command(self):
        ''' Exec Command '''
        now = int(time.time())

        queue_base = self.queue_base
        queue_key = self.queue_key

        tx = self.client.pipeline(transaction=True)
        tx.hmget(queue_key, "vt", "delay", "maxsize", "totalrecv", "totalsent", "created",
                 "modified")
        tx.zcard(queue_base)
        tx.zcount(queue_base, "%s000" % now, "+inf")

        results = tx.execute()

        if not results or results[0][0] is None:
            raise QueueDoesNotExist(self.get_qname)

        stats = results[0]

        return {
            "vt": stats[0],
            "delay": stats[1],
            "maxsize": stats[2],
            "totalrecv": stats[3] or 0,
            "totalsent": stats[4] or 0,
            "created": stats[5],
            "modified": stats[6],
            "msgs": results[1],
            "hiddenmsgs": results[2]
        }
